/**
 * Copyright (c) Members of the EGEE Collaboration. 2004-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2009-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     Mischa Sall\'e <msalle@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *  2007-2009
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 *  2003-2007
 *     Martijn Steenbakkers <martijn@nikhef.nl>
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *
 */

/*!
    \file   lcas_verify-proxy.c
    \brief  Interface to the LCAS plugins
    \author Oscar Koeroo for the EGEE project

    LCAS plugin that makes authorization decisions based the verification of the proxy and its X.509 based chain
    -# plugin_initialize()
    -# plugin_confirm_authorization()
    -# plugin_terminate()
*/

/******************************************************************************

lcas_check-executable.c

Description:
    LCAS module that checks if the executable name found in the rsl is the allowed executable or not.

CVS Information:
    $Source: /srv/home/dennisvd/svn/mw-security/lcas-plugins-check-executable/src/check-executable/lcas_check-executable.c,v $
    $Date: 2010-05-03 09:05:22 $
    $Revision: 1.12 $
    $Author: okoeroo $

******************************************************************************/

/*****************************************************************************
                            Include header files
******************************************************************************/
#include "lcas_config.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if HAVE_MALLOC_H
#include <malloc.h>
#endif

#if HAVE_TIME_H
#include <time.h>
#endif

#if HAVE_VALUES_H
#include <values.h>
#endif

#include <lcas/lcas_modules.h>

/******************************************************************************
                                Definitions
******************************************************************************/

/*
 */

/******************************************************************************
                          Module specific prototypes
******************************************************************************/

char * getPos (char * src, char * pattern, int * offset);
int searchRSLexe (char * rsl, char ** exe);

/******************************************************************************
                       Define module specific variables
******************************************************************************/

static char *       modname         =  "lcas_check-executable.mod";
static char *       allowedExecList =  NULL;

/******************************************************************************
Function:   plugin_initialize
Description:
    Initialize plugin
Parameters:
    argc, argv
    argv[1]: database to be used by plugin
Returns:
    LCAS_MOD_SUCCESS : succes
    LCAS_MOD_FAIL    : failure
******************************************************************************/
#if 0
int plugin_initialize(char * plugin_db)
#endif
int plugin_initialize(int argc, char ** argv)
{
    int i = 0;
    int gotExec = 0;

    lcas_log_debug(1,"%s-plugin_initialize(): passed arguments:\n",modname);
    for (i=0; i < argc; i++)
    {
        lcas_log_debug(1,"%s-plugin_initialize(): arg %d is %s\n",
             modname,i,argv[i]);
    }

    for (i = 1; i < argc; i++)
    {
        if ( ((strcmp(argv[i], "-exec") == 0) ||
              (strcmp(argv[i], "-EXEC") == 0))
             && (i + 1 < argc))
        {
            if (gotExec)
            { 
                lcas_log (0,"%s: Error in initialization parameter: %s (failure) - parameter has already been used once. Use: -exec <executable[:executable[:executable[...]]]>\n", modname, argv[i]);
                return LCAS_MOD_FAIL;
            }

            else 
            {
                if ((argv[i + 1] != NULL) && (strlen(argv[i + 1]) > 0))
                {
                    allowedExecList = strdup(argv[i + 1]);
                    gotExec = 1; // -exec is found and handle
                }
                i++;

               continue;
            }
        }

        else
        {
            lcas_log (0,"%s: Error in initialization parameter: %s (failure) - module only allows -exec <path[:path]*> \n", modname, argv[i]);
            return LCAS_MOD_FAIL;
        }
    }

    return LCAS_MOD_SUCCESS;
}

/******************************************************************************
Function:   plugin_confirm_authorization
Description:
    Ask for authorization by passing RSL and user credential
Parameters:
    request:   RSL request
    user_cred: user credential
Returns:
    LCAS_MOD_SUCCESS : succes
    LCAS_MOD_FAIL    : failure
******************************************************************************/
int
plugin_confirm_authorization(lcas_request_t request, lcas_cred_id_t lcas_cred)
{
    int                           passflag = 0;
    char                        * rsl = request;
    char                        * f = NULL;
    int                           i = 0;
    char                       ** list = NULL;
    int                           eCount = 0;


    lcas_log_debug(1,"%s-plugin_confirm_authorization: Entering check-executable plugin\n", modname);
    lcas_log_debug(2,"%s-plugin_confirm_authorization: RSL: %s\n", modname, request);
    // search the rsl for the executable
    if (searchRSLexe(rsl, &f))
    {
        lcas_log_debug(4,"            f = %s & allowedExecList = %s\n", f, allowedExecList);

        // split the input arg list and can't be empty
        if (splitList (allowedExecList, &list, &eCount) && list != NULL)
        {
            lcas_log_debug(4,"                eCount = %d & list is located at: %d\n", eCount, list);

            // Walk through the list of eCount elements
            for (i = 0; i < eCount; i++)
            {
                lcas_log_debug(4,"                list[%d] = %s\n", i, list[i]);
                
                if (passflag = ((strcmp(f, list[i]) == 0) ? 1 : 0))
                    break;
            }
        }
        else
        {
            lcas_log_debug(0,"%s-plugin_confirm_authorization: Couldn't split the input arguments correctly: \"%s\"\n", modname, allowedExecList);
            passflag = 0;
        }
    }
    else
    {
        lcas_log_debug(0,"%s-plugin_confirm_authorization: No executable found in the RSL\n", modname);
        passflag = 0;
    }


    // Need to do clean up of created structures!
    if (f != NULL) free (f);

    /* Only cleanup first element, as they are all malloced combined using a
     * strdup */
    if (list != NULL)	{
	free(list[0]);	list[0]=NULL;
	free(list);	list=NULL;
    }

    // Final Evaluation
    if (passflag != 1) return LCAS_MOD_FAIL;

    return LCAS_MOD_SUCCESS;
}

/******************************************************************************
Function:   plugin_terminate
Description:
    Terminate plugin
Parameters:

Returns:
    LCAS_MOD_SUCCESS : succes
    LCAS_MOD_FAIL    : failure
******************************************************************************/
int plugin_terminate()
{
    lcas_log_debug(1,"%s-plugin_terminate(): terminating\n",modname);

    return LCAS_MOD_SUCCESS;
}



/******************************************************************************
Function:   splitList
Description:
    Will create a array of char * to hold each element from the src string.
    delimited by ':'
Parameters:
    char *    src     : source string
    char ***  retList : return List with elements
    int  *    count_elements : return value for the amount of elements in the list.
Returns:
    1        : succes
    0        : failure
******************************************************************************/

int splitList (char * src, char *** retList, int * count_elements)
{
    char ** temp_list = NULL;
    int     e    = 0;
    int     i    = 0;
    int len,count= 1;
    char *dummy  = NULL;

    /* initialize returned values */
    *count_elements=0;
    *retList=NULL;

    /* Protect against strlen(NULL) */
    if ( src==NULL || (len=strlen(src))==0 )
	return 0;

    /* Duplicate input */
    if ( (dummy=strdup(src))==NULL )	{
	lcas_log_debug (0, "%s-%s(): out of memory\n", modname, __func__);
	return 0;
    }

    /* calc the elementcount */
    for (i = 0; i < len; i++) {
        if (src[i] == ':')
	    count++;
    }

    /* Allocate the amount of elements in array && Preallocate the amount of
     * buffer space per element */
    if ( (temp_list = calloc (count, sizeof (char *))) == NULL )	{
	free(dummy);
	lcas_log_debug (0, "%s-%s(): out of memory\n", modname,__func__);
	return 0;
    }

    /* First entry is start of dummy */
    temp_list[e++]=dummy;
    
    /* Go over other entries */
    for (i=0; i<len; i++)   {
	if (dummy[i]==':')	{
	    temp_list[e++]=&(dummy[i+1]);
	    dummy[i]='\0';
	}
    }

    /* Log output */
    for (i = 0; i < count; i++)
        lcas_log_debug(4,"        %s: fetched: %s\n", __func__, temp_list[i]);

    /* Return the values */
    *count_elements=count;
    *retList = temp_list;
    return 1;
}


/******************************************************************************
Function:   getPos
Description:
    returns a char pointer to the location where pattern is found or NULL
Parameters:
    char * src     : source string to look for pattern
    char * pattern : search pattern to look for in src
    int  * offset  : return value for the offset where pattern is found or -1 for not finding anything
Returns:
    Pointer to char in the src string : succes
    NULL                              : failure
******************************************************************************/
char * getPos (char * src, char * pattern, int * offset)
{
   // init
   int c = 0;
   char * r = NULL;
   char * t = src;
   int l = strlen(src);
  
   *offset = 0;

   // search for pattern
   while (t != NULL && t != '\0')   
   {
      // Searching...
      if (strncmp (t, pattern, strlen(pattern)) != 0)
      {      
         // Next
         t = &t[1];
      }
      else
      {
         // Gevonden!
         *offset = c;
         r = t;
         break;         
      }

      // offset counter
      c++;
   }
  
   return r;
}


/******************************************************************************
Function:   searchRSLexe
Description:
    Searches the rsl for the executable field and return its value in char ** exe
Parameters:
    char *  rsl : the rsl string
    char ** exe : pointer to the malloced space which holds the executable string

    NOTE : the char ** exe must be non-allucated, I'll perform that action on the right spot at the right time overhere
Returns:
    0  : succes
    !0 : failure (not found)
******************************************************************************/
int searchRSLexe (char * rsl, char ** exe)
{
   int     os = 0;    // OffSet
   char *  p  = NULL;

   // Let p point to beginning of pattern
   if (p = getPos(rsl, "\\\"executable\\\"", &os))
   {
      // Let p point to beginning of pattern
      if (p = getPos(p, "=", &os))
      {
         // Let p point to beginning of pattern
         if (p = getPos(p, "\\\"", &os))
         {
            // Check if the p (thus the rsl) is still big enough
            if (strlen(p) > 2)
            {
               // Check from one position up the string (the start of the executable string) what the offset is of the closing quote of the executable
               if (getPos((&p[2]), "\\\"", &os))
               {
                  // Holy Smokes, I've found it all, let's extract the executable string to char ** exe
                  *exe = calloc (os + 1, sizeof(char));
                  strncpy (*exe, &p[2], os);
               }
               else
               { 
                  // couldn't find the closing quote, bad rsl
                  lcas_log_debug(0,"%s-searchRSLexe: couldn't find the closing quote, bad rsl\n",modname);
                  return 0;
               }
            }
            else
            {
               // End of string reached -> failure
               lcas_log_debug(0,"%s-searchRSLexe: end of string reached, after first quote found, bad rsl\n",modname);
               return 0;
            }
         }
         else
         {
            // Couldn't extract: '\\\"', bad RSL
            lcas_log_debug(0,"%s-searchRSLexe: couldn't find the '\\\"' character surrounding the executable string, bad rsl\n",modname);
            return 0;
         }
      }
      else
      {
         // Couldn't extract: '=', bad RSL
         lcas_log_debug(0,"%s-searchRSLexe: couldn't find the '=' character between 'executable' and its value, bad rsl\n",modname);
         return 0;
      }
   }
   else
   {
      // Couldn't extract: "\"executable\"", bad RSL
      lcas_log_debug(0,"%s-searchRSLexe: couldn't find the '\"\"executable\"\"', bad rsl\n",modname);
      return 0;
   }

   return 1;
}
